-- This file is a part of Falling Block Game 2.
-- 
-- Copyright (C) 2011 Jared Krinke.
-- 
-- Falling Block Game 2 is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

require("Fbg2/Common.lua")
require("Effects/Path.lua")
require("Utility/MiscSettings.lua")

Piece = { }

function Piece.new(value, configs)
    local piece = Entity.new()

    piece.value = value
    piece.configs = configs

    piece.horizontalOffset = Path.new()
    piece.verticalOffset = Path.new()
    piece.angleOffset = Path.new()

    function piece.update(self, ms)
        if Fbg2.getAnimate() then
            -- Update smooth drop period (in case it has changed)
            local board = self.parent

            if board ~= nil then
                self.verticalOffset:setPeriod(board:getDropPeriod())
            end

            -- Update paths
            self.horizontalOffset:update(ms)
            self.verticalOffset:update(ms)
            self.angle = self.angleOffset:update(ms)

            -- Update position (taking offsets into account)
            self:updatePosition()
        end
    end

    function piece.updatePosition(self)
        -- Update x and y coordinates, taking into account config's center
        if isNumber(self.i) and isNumber(self.j) then
            -- Entity's offset is zero-based, but coordinates are one-based and blocks are centered
            self.x = self.i - 0.5 + self.config.centerOffsetX + self.horizontalOffset:getValue()
            self.y = self.j - 0.5 + self.config.centerOffsetY + self.verticalOffset:getValue()
        end
    end

    function piece.setConfigIndex(self, i)
        self.configIndex = i
        self.config = self.configs[i]
        self.elements = self.config.elements

        self:updatePosition()
        self.angle = self.angleOffset:setValue((i - 1) * -90)
    end

    function piece.reset(self, board, i, j)
        self.horizontalOffset:setValue(0)
        self.verticalOffset:setValue(0)
        self.angleOffset:setValue(0)

        self:setPosition(i, j)
        self:setConfigIndex(1)
    end

    function piece.getPosition(self)
        return self.i, self.j
    end

    function piece.setPosition(self, i, j)
        self.i = i
        self.j = j
        self:updatePosition()
    end

    function piece.forEachBlock(self, config, callback, testI, testJ)
        local i0, j0 = self:getPosition()
        local continueIteration = true

        if isNil(config) then
            config = self.config
        end

        if isNumber(testI) and isNumber(testJ) then
            i0 = testI
            j0 = testJ
        end

        local c = 1

        while continueIteration and c <= #config do
            local configBlocks = config[c]
            local i = i0 + configBlocks[1]
            local j = j0 + configBlocks[2]

            continueIteration = callback(i, j)

            c = c + 1
        end

        return valid
    end

    function piece.validate(self, board, config, testI, testJ)
        local getBlockOccupied = board.getBlockOccupied
        local width = board:getWidth()
        local height = board:getHeight()
        local valid = true

        self:forEachBlock(config, function(i, j)
            -- Check boundaries (note that overflowing the vertical bound is acceptable)
            if i >= 1 and i <= width and j >= 1 then
                -- Check to make sure the block is not occupied
                if j <= height and getBlockOccupied(board, i, j) then
                    valid = false
                end
            else
                valid = false
            end

            return valid
        end, testI, testJ)

        return valid
    end

    function piece.tryMove(self, board, offsetI, offsetJ)
        local i, j = self:getPosition()
        local moved = false

        i = i + offsetI
        j = j + offsetJ

        if self:validate(board, self.config, i, j) then
            -- Set the new position
            self:setPosition(i, j)
            moved = true

            -- Smooth drop animation, if necessary
            if Fbg2.getAnimate() then
                if not self:validate(board, self.config, i, j + 1) then
                    self.verticalOffset:setValue(0)
                elseif offsetJ ~= 0 then
                    self.verticalOffset:schedule(-offsetJ, 0, board:getDropPeriod())
                end
            end
        end

        return moved
    end

    function piece.tryRotate(self, board, configIndexOffset)
        local oldConfig = self.config
        local i = (self.configIndex - 1 + configIndexOffset) % #self.configs
        local rotated = false

        if i < 0 then
            i = #self.configs + i + 1
        end

        if self:validate(board, self.configs[i + 1]) then
            self:setConfigIndex(i + 1)
            rotated = true

            -- Rotate animation, if necessary
            if Fbg2.getAnimate() then
                self.angleOffset:schedule(self.angle + 90 * configIndexOffset, self.angle, 67)
                self.horizontalOffset:schedule(oldConfig.centerOffsetX - self.config.centerOffsetX, 0, 67)
            end
        end

        return rotated
    end

    function piece.complete(self, board)
        local config = self.config
        local width = board:getWidth()
        local height = board:getHeight()
        local i0, j0 = self:getPosition()
        local c = 1

        while c <= #config do
            local configBlocks = config[c]
            local i = i0 + configBlocks[1]
            local j = j0 + configBlocks[2]

            if i >= 1 and i <= width and j >= 1 and j<= height then
                board:setBlock(i, j, self.value)
            end

            c = c + 1
        end
    end

    function piece.tryMoveLeft(self, board)
        return self:tryMove(board, -1, 0)
    end

    function piece.tryMoveRight(self, board)
        return self:tryMove(board, 1, 0)
    end

    function piece.tryMoveDown(self, board)
        return self:tryMove(board, 0, -1)
    end

    function piece.tryRotateClockwise(self, board)
        return self:tryRotate(board, 1)
    end

    function piece.tryRotateCounterclockwise(self, board)
        return self:tryRotate(board, -1)
    end

    local directions = { -1, 1 }

    function piece.processCloudEffects(self, board, process)
        local getBlockOccupied = board.getBlockOccupied
        local width = board:getWidth()
        local height = board:getHeight()
        local d = 1

        while d <= #directions do
            local direction = directions[d]
            local lowerBounds = { }

            -- Find lower bounds
            self:forEachBlock(self.config, function(i, j)
                if i >= 1 and i <= width and j >= 1 and j <= height then
                    if isNil(lowerBounds[i]) or j < lowerBounds[i] then
                        lowerBounds[i] = j
                    end
                end

                return true
            end)

            -- Check lower bounds
            Table.forEach(lowerBounds, function(i, j)
                local i2 = i + direction
                
                if i2 >= 1 and i2 <= width and j >= 1 and j <= height then
                    -- Check horizontal direction (should be empty)
                    if not getBlockOccupied(board, i2, j) then
                        -- Check vertical direction
                        if j <= 1 or getBlockOccupied(board, i, j - 1) then
                            -- Float up if there's a block below this space
                            local floatUp = j <= 1 or getBlockOccupied(board, i + direction, j - 1)

                            process(i, j, direction, floatUp)
                        end
                    end
                end
            end)

            d = d + 1
        end

        return valid
    end

    -- Set up center for each config
    local i = 1

    while i <= #configs do
        local config = configs[i]
        local j = 1

        -- Determine config's dimensions for centering
        local minX = 5
        local maxX = -5
        local minY = 5
        local maxY = -5

        while j <= #config do
            local configBlocks = config[j]

            minX = min(minX, configBlocks[1])
            maxX = max(maxX, configBlocks[1])
            minY = min(minY, configBlocks[2])
            maxY = max(maxY, configBlocks[2])

            j = j + 1
        end

        -- Offset of the center
        config.centerOffsetX = minX + (maxX - minX) / 2
        config.centerOffsetY = minY + (maxY - minY) / 2

        -- Setup elements (each config needs its own elements since otherwise
        -- blocks--with their shadows--would actually be rotated/inconsistent
        local configElements = ElementList.new()

        j = 1

        while j <= #configs[1] do
            -- Always use the first config's elements since the parent will already be rotated/positioned
            local configBlocks = configs[1][j]
            local image = Element.Image.new(Fbg2.blockImages[value + 1], configBlocks[1] - configs[1].centerOffsetX, configBlocks[2] - configs[1].centerOffsetY)

            image.angle = (i - 1) * 90
            configElements:add(image)
            j = j + 1
        end

        config.elements = configElements
        i = i + 1
    end

    -- Set first config
    piece:setConfigIndex(1)

    return piece
end

