-- This file is a part of Falling Block Game 2.
-- 
-- Copyright (C) 2011 Jared Krinke.
-- 
-- Falling Block Game 2 is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
-- 
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
-- 
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

require("Utility/SortedList.lua")
require("Fbg2/Common.lua")
require("Settings/HighScoreDefaultData.lua")
include("Settings/HighScoreData.lua")

HighScore = { }

HighScore.count = 10

function HighScore.compare(a, b)
    local v = 0

    if a.score > b.score then
        v = -1
    elseif a.score < b.score then
        v = 1
    end

    return v
end

-- Mapping from mode (and any other parameters) to the high score page
local highScorePageNames = {
    "Classic",
    "Cleanup (Height 0)",
    "Cleanup (Height 1)",
    "Cleanup (Height 2)",
    "Cleanup (Height 3)",
    "Cleanup (Height 4)",
    "Cleanup (Height 5)",
}

local modeToGetPage = {
    -- Page 1 is Classic mode
    [Fbg2.modes.Classic] = function() return 1 end,

    -- Pages 2-7 are for Cleanup mode (with different heights)
    [Fbg2.modes.Cleanup] = function(height) return 2 + height end,
}

function HighScore.getPage(mode, height)
    local getPage = modeToGetPage[mode]
    local page = nil

    if isFunction(getPage) then
        page = getPage(height)
    end

    return page
end

function HighScore.getPageName(page)
    return highScorePageNames[page]
end

function HighScore.getPageCount()
    return #highScorePageNames
end

function HighScore.new(name, score, cleared, level)
    return {
        name = name,
        score = score,
        cleared = cleared,
        level = level,
    }
end

function HighScore.getLowestScore(mode, height)
    local page = HighScore.getPage(mode, height)
    local lowestScore = 0

    if isNumber(page) then
        local list = highScoresLists[page]

        if isTable(list) then
            local lastScore = list:getLast()

            if not isNil(lastScore) then
                lowestScore = lastScore.score
            end
        end
    end

    return lowestScore
end

local function insertIntoPage(page, highScore)
    if isNumber(page) then
        local list = highScoresLists[page]

        if isTable(list) then
            list:insert(highScore)
            list:truncate(HighScore.count)
        end
    end
end

function HighScore.insert(mode, height, highScore)
    local page = HighScore.getPage(mode, height)

    insertIntoPage(page, highScore)
end

function HighScore.save()
    File.createDirectory("Settings")

    local file = File.openWrite("Settings/HighScoreData.lua")
    local page = 1

    while page <= #highScorePageNames do
        local i = 1

        highScoresLists[page]:forEach(function(highScore)
            file:write(String.format("highScoreDataV2[%d][%d] = { %q, %d, %d, %d }", page, i, highScore.name, highScore.score, highScore.cleared, highScore.level))
            i = i + 1
        end)

        page = page + 1
    end

    file:close()
end

-- Set up highScoresLists
local i = 1
highScoresLists = { }

while i <= #highScorePageNames do
    highScoresLists[i] = SortedList.new(HighScore.compare)
    i = i + 1
end

-- Load high score data
i = 1

while i <= #highScorePageNames and not isNil(highScoreDataV2[i]) do
    local highScores = highScoreDataV2[i]
    local j = 1

    while not isNil(highScores[j]) do
        local highScore = highScores[j]

        insertIntoPage(i, HighScore.new(highScore[1], highScore[2], highScore[3], highScore[4]))
        j = j + 1
    end

    i = i + 1
end

